FONTS_AVAILABLE = [
  'Acme',
  'Arial',
  'Arvo',
  'Avenir Next',
  'Bangers',
  'Cabin',
  'Cabin Condensed',
  'Courier',
  'Crete Round',
  'Dancing Script',
  'Exo',
  'Georgia',
  'Handlee',
  'Helvetica Neue',
  'Karla',
  'Lato',
  'Lekton',
  'Lobster',
  'Lora',
  'McLaren',
  'Montserrat',
  'Nixie One',
  'Old Standard TT',
  'Open Sans',
  'Oswald',
  'Playfair Display',
  'Quicksand',
  'Raleway',
  'Signika',
  'Sniglet',
  'Source Sans Pro',
  'Vollkorn'
]

API_BASE_URL = 'https://api.typeform.com'
