### CLASS TO DEAL WITH MICROSOFT CMD TERMINAL ###

class BColors:
    '''
    REFERENCES: https://stackoverflow.com/questions/4406389/if-else-in-a-list-comprehension
    DOCSTRING:
    INPUTS:
    OUTPUTS
    '''
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
